/* Nessus
 * Copyright (C) 1998 - 2003 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * In addition, as a special exception, Renaud Deraison
 * gives permission to link the code of this program with any
 * version of the OpenSSL library which is distributed under a
 * license identical to that listed in the included COPYING.OpenSSL
 * file, and distribute linked combinations including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * this file, you may extend this exception to your version of the
 * file, but you are not obligated to do so.  If you do not wish to
 * do so, delete this exception statement from your version.
 *
 * pluginscheduler.c : 
 *  Tells nessusd which plugin should be executed now
 *
 */
 
#include <includes.h>
#define IN_SCHEDULER_CODE 1
#include "pluginscheduler.h"
#include "pluginload.h"
#include "pluginlaunch.h"




/*---------------------------------------------------------------------------*
 *
 * A minimalist HASH stucture
 *
 *---------------------------------------------------------------------------*/
#define HASH_MAX 2713


	
static int mkhash(char * name)
{
  int l = strlen(name);
  int i;
  int h = 0;
  
  for(i=0;i<l;i++)
   h = ((h * HASH_MAX) + name[i]) % HASH_MAX;
   
 
  return h < 0 ? -h:h;
}	

static struct hash * hash_init()
{
 struct hash * h = emalloc(sizeof(*h) * HASH_MAX + 1);

 return h;
}

static void hash_link_destroy(struct hash * h)
{
 if(h == NULL)
  return;
  
 if(h->next != NULL)
  hash_link_destroy(h->next);
 efree(&h);
}

static void hash_destroy(struct hash * h)
{
 int i;
 
 for(i=0;i<HASH_MAX;i++)
  {
  hash_link_destroy(h[i].next);
  }
 efree(&h);
}


static int hash_add(struct hash * h, char * name, struct arglist * plugin)
{
 struct hash * l = emalloc(sizeof(struct hash));
 int idx = mkhash(name);
 
 l->plugin = plugin;
 l->name   = name;
 l->next = h[idx].next;
 h[idx].next = l;

 return 0;
}


static struct arglist * hash_get(struct hash * h, char * name)
{
 int idx = mkhash(name);
 struct hash * l = h[idx].next;
 while(l != NULL)
 {
  if(strcmp(l->name, name) == 0)
   return l->plugin;
  else
   l = l->next;
 }
 return NULL;
}

/*----------------------------------------------------------------------*
 *									*
 * Utilities								*
 *									*
 *----------------------------------------------------------------------*/



/*
 * Returns the category of <plugin>
 */
int plugin_get_category(struct arglist * plugin)
{
 struct arglist * a;
 int ret;
 
 if(plugin == NULL)
  return -1;
 
 a = arg_get_value(plugin, "plugin_args");
 if(a == NULL)
  return -1;
 
 ret = (int)arg_get_value(a, "CATEGORY");
 return ret;
}


/*
 * Returns the dependencies of <plugin>
 */
struct arglist * plugin_get_dependencies(struct arglist * plugin)
{
 struct arglist * a, * b;
 
 
 if(plugin == NULL)
  return NULL;
  
 a = arg_get_value(plugin, "plugin_args");
 if(a == NULL)
  return NULL;
 
 b = arg_get_value(a, "DEPENDENCIES");
 
 
 return b;
}


int plugin_get_running_state(struct arglist * plugin)
{
 return (int)arg_get_value(plugin, "RUNNING_STATE");
}

void plugin_set_running_state(struct arglist * plugin, int state)
{
 if(plugin == NULL)
  return;
  
 if(arg_get_value(plugin, "RUNNING_STATE") != NULL)
  arg_set_value(plugin, "RUNNING_STATE", sizeof(state), (void*)state);
 else 
  arg_add_value(plugin, "RUNNING_STATE", ARG_INT, sizeof(state), (void*)state);
}



struct arglist * plugin_next_unrun_dependencie(struct hash * h, struct arglist * dependencies)
{
 int flag = 0;
 if(dependencies == NULL)
  return NULL;
  
  while(dependencies->next)
  {
   struct arglist * plugin = hash_get(h, dependencies->name);
   if(plugin != NULL)
   {
    int state = plugin_get_running_state(plugin->value);
    switch(state)
    {
     case PLUGIN_STATUS_UNRUN :
     	{
	struct arglist * deps = plugin_get_dependencies(plugin->value);
	struct arglist * ret;
	if(deps == NULL)
     		return plugin;
	ret = plugin_next_unrun_dependencie(h, deps);
	if(ret == NULL)
		return plugin;
	else 
		return ret;
	}
     case PLUGIN_STATUS_RUNNING:
     	flag++;
	break;
     case PLUGIN_STATUS_DONE:
     	break;
    }
   }
   else fprintf(stderr, "%s could not be found\n", dependencies->name);
   dependencies = dependencies->next;
  }
  
  
  if(flag == 0)
  	return NULL;
  else
  	return PLUG_RUNNING;
}

/*---------------------------------------------------------------------------*/

/*
 * Enables a plugin and its dependencies
 */
static void enable_plugin_and_dependencies(plugins_scheduler_t shed, struct arglist * plugin, char * name)
{
 struct arglist * deps = plugin_get_dependencies(plugin);
 

 plug_set_launch(arg_get_value(plugin, "plugin_args"), 1);
 
 if(deps != NULL)
 {
   while(deps->next != NULL)
    {
     struct arglist * p;
     p = hash_get(shed->hash, deps->name);
     if( p != NULL )
       enable_plugin_and_dependencies(shed, p->value, p->name);
     else
       fprintf(stderr, "'%s' depends on '%s' which could not be found\n", name, deps->name);
     
     deps = deps->next;
    }
 }
}

/*---------------------------------------------------------------------------*/

plugins_scheduler_t plugins_scheduler_init(struct arglist * plugins, int autoload, int simul)
{
 plugins_scheduler_t ret = emalloc(sizeof(*ret));
 struct arglist * arg;
 int i;
 
 
 
 
 if(plugins == NULL)
  return NULL;
 
 
 /*
  * Fill our list in reverese order of family, to minimize
  * our exploration.
  */
 for(i = ACT_LAST;i>=ACT_FIRST;i--)
 {
  arg = plugins;
  while(arg->next != NULL)
  {
  struct list * dup;
  struct arglist * args = arg_get_value(arg->value, "plugin_args");
  
  if(i == (int)arg_get_value(args, "CATEGORY"))
  {
  dup = emalloc(sizeof(struct list));
  dup->name = arg->name;
  dup->plugin = arg;
  dup->prev = NULL;
  dup->next = ret->list;
  if(ret->list != NULL)
   ret->list->prev = dup;
  ret->list = dup;
  }
  arg = arg->next;
  } 
 }
 
 
 ret->hash = hash_init();
 arg = plugins;
 while(arg->next != NULL)
 {
  hash_add(ret->hash, arg->name, arg);
  plugin_set_running_state(arg->value, PLUGIN_STATUS_UNRUN);
  arg = arg->next;
 }
 
 
 if(autoload != 0)
 {
 arg = plugins;
 while(arg->next != NULL)
  {
   if(plug_get_launch(arg_get_value(arg->value, "plugin_args")) != 0)
   	enable_plugin_and_dependencies(ret, arg->value, arg->name);
   arg = arg->next;
  }
 }
 
 ret->simul = simul;
 return ret;
}



struct arglist * plugins_scheduler_next(plugins_scheduler_t h)
{
 
 struct list * l;
 int category;
 int flag = 0;
 
 if(h == NULL)
  return NULL;
 
 for(category = ACT_FIRST;category<=ACT_LAST;category++)
 {

 l = h->list;
 
 /*
  * Scanners (and DoS) must not be run in parrallel
  */
 if(h->simul == 0)
 {
 if((category == ACT_SCANNER) ||
    (category == ACT_KILL_HOST) ||
    (category == ACT_DENIAL))
    pluginlaunch_disable_parrallel_checks();
 else
    pluginlaunch_enable_parrallel_checks();
 }  
      
 while(l != NULL)
 {
   int state;

   if(category != plugin_get_category(l->plugin->value))
   {
    l = l->next;
    continue;
   }
  
  state = plugin_get_running_state(l->plugin->value);
  
  switch(state)
  {
   case PLUGIN_STATUS_UNRUN:
    {
    struct arglist * deps = plugin_get_dependencies(l->plugin->value);
    if(deps != NULL)
    {
     struct arglist * p = plugin_next_unrun_dependencie(h->hash, deps);
     switch((int)p)
     {
      case (int)NULL :
	 plugin_set_running_state(l->plugin->value, PLUGIN_STATUS_RUNNING);
	 return l->plugin;
     case (int)PLUG_RUNNING:
     		/* One of the dependencie is still running */
     	flag ++;
	break;
     default:
        { 
	 /* Launch a dependencie */
	 int c;
	 
	 /* Ack - the plugin we depend on is not of the same category - we recast the current
	  * plugin to another (greater) category 
	  */
	 if((c = plugin_get_category(p->value)) != category)
	 {
	  struct arglist * args = arg_get_value(l->plugin->value, "plugin_args");
	  arg_set_value(args, "CATEGORY", sizeof(int), (void*)c);
	  l = l->next;
	  continue;
	 }
	 
	 plugin_set_running_state(p->value, PLUGIN_STATUS_RUNNING);
	 return p;
	 }
      }
    }
    else /* No dependencies */
     {
      plugin_set_running_state(l->plugin->value, PLUGIN_STATUS_RUNNING);
      return l->plugin;
     }
   }
    break;
  case PLUGIN_STATUS_RUNNING:
  	flag ++;
	break;

  case PLUGIN_STATUS_DONE:
  	{
	 struct list * old = l->next;
	 
	

	 if(l->prev != NULL)
	  l->prev->next = l->next;
	 else
	  h->list = l->next;
	  
	 
	 if(l->next != NULL)
	  l->next->prev = l->prev;

	efree(&l);
	l = old;
	
	continue;
	}
	break;

  }
  l = l->next; 
  }
  
  
  /* Could not find anything */
  if((category == ACT_SCANNER ||
     category == ACT_INIT) && flag != 0)
     {
      if(h->simul == 0)pluginlaunch_wait_for_free_process();
      flag = 0;
      category --;
     }
 }
 
   return flag == 0 ? NULL : PLUG_RUNNING; 
}


void list_destroy(struct list * list)
{
 while(list != NULL)
 {
  struct list * next = list->next;
  efree(&list);
  list = next;
 }
}


void plugins_scheduler_free(plugins_scheduler_t sched)
{
 hash_destroy(sched->hash);
 list_destroy(sched->list);
 efree(&sched);
}
